# MIQCP - maximise ANP
import pandas as pd
import numpy as np
import os
import numpy.random
from numpy import arange
import xlwt
from openpyxl import load_workbook
import mpmath
import pyomo.environ as pe
import pyomo.opt as po
from pyomo.opt import SolverFactory
from pyomo.core.base.PyomoModel import AbstractModel
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.set import RangeSet
import gurobipy
from pyomo.environ import *

# locate file and load the workbook
directory = "/home/jessieluo"
os.chdir(directory)
file_name = "process results.xlsm"
file = directory + "/" + file_name
wb1 = load_workbook(file_name, keep_vba=True, data_only=True)  # Work Book
wb = load_workbook(file, keep_vba=True)

# plant info
min_operatingload = 0.70  # based on the volume flexibility results at the equipment level
ws_elec = wb1['Overall electricity balance']
MLload_100 = ws_elec['A27'].value  # nominal electricity consumption, GJ/h
ws_economics = wb1["CAPEX & OPEX"]
ws_ecoassumption = wb1['Economic assumptions'] # a hidden spreadsheet
lifetime = ws_ecoassumption['B7'].value  # life time, y
discount_rate = ws_ecoassumption['B9'].value
salvage_rate = ws_ecoassumption['B10'].value
projection_2030 = 1  # from NREL
storage_info = {"type": ["Li", "Lead", "RFB"],
                "unit_cost": [1200/3.6/1000*projection_2030, 500/3.6/1000*projection_2030, 1000/3.6/1000*projection_2030],
                "efficiency": [0.9, 0.85, 0.85],
                "salvage_rate": [0.05, 0.05, 0.05]} # unit cost: M€/GJ
# read CAPEX and OPEX parameters
# It and OMt are fixed at the 100% load
It_ML_ref = [0 for i in range(lifetime+1)] # CAPEX flow, M€/y
It_ML_ref[0] = ws_economics['G18'].value  # specify the initial CAPEX, M€/y
It_ML_ref[15] = (ws_ecoassumption['M59'].value + ws_ecoassumption['N59'].value) / 1e6  # specify the replacement of the membrane contactor
It_ML_ref[30] = - It_ML_ref[0] * salvage_rate  # specify the salvage value M€
OMt_ML_load100 = ws_economics['G22'].value # operation and maintenance cost at 100% load, M€/y


# Wt, Ft, Et are always linear to LR, do need to be adjusted based on power profile
ws_mass = wb1['Overall mass balance']
Wt_ML_load100 = ws_mass['B57'].value  # waste treatment cost at 100% load, €/h
Ft_ML_load100 = ws_mass['B56'].value  # feedstock cost at 100% load, €/h
Et_ML_load100 = ws_elec['A31'].value  # electricity cost at 100% load, €/h
Mt_ML_load100 = ws_mass['B65'].value  # production of C6A at 100% load, kg/h

# For this optimisation, Ht is linearised to LR, €/h
ws_heat = wb1["Overall utility balance"]
Ht_ML_load100 = ws_heat['B81'].value # heat cost at 100% load, €/h

# economic allocation factor of C6A
C6A_AF = ws_mass['B63'].value

# read power profile and price
ws_elecprofile = wb1["Electricity profile & pricing"]
hybridpower_column = ws_elecprofile['O']
hybridpower_list = [round(hybridpower_column[x].value, 2) for x in range(2, 8762)] # LR of the hybrid profile, %
hybridpowerprice_column = ws_windprofile['Q']
hybridpower_price = [hybridpowerprice_column[x].value for x in range(2, 8762)] # price of hybrid electricity, €/GJ


# select operating cases
power_list = hybridpower_list  # select between hybridpower_list and windpower_list
load_100 = 1  # nominal load ratio
convert = 64  # m3/h

# how many hours retrieved
lower = 0
upper = 8760
T = range(lower, upper, 1)

# start
model = pe.ConcreteModel()

# sets
model.N = pe.Set(initialize=T)

# initialise parameters
model.power = pe.Param(model.N, initialize=power_list[lower:upper]) # load ratio, %
model.E = pe.Param(initialize=Et_ML_load100)  # €/h
model.F = pe.Param(initialize=Ft_ML_load100)  # €/h
model.WT = pe.Param(initialize=Wt_ML_load100)  # €/h
model.H = pe.Param(initialize=Ht_ML_load100)  # €/h

# initialise variables
model.M_li = pe.Var(model.N, domain=pe.NonNegativeReals) # m3/h
model.M_lo = pe.Var(model.N, domain=pe.NonNegativeReals) # m3/h
model.ST = pe.Var(domain=pe.NonNegativeReals, bounds=(33, 43)) # m3

# one binary variable and one continuous variable
model.L = 0.70
model.production_select = pe.Var(model.N, domain=pe.Binary)
model.production = pe.Var(model.N, domain=pe.NonNegativeReals, bounds=(0.00, 1.00)) # ratio to nominal production, %

# constants
Mt_ML_load100 = Mt_ML_load100/1e3  #t/h
nominal_production_convert = 1 / convert * Mt_ML_load100  # t/m3
nominal_load_convert = 1 / MLload_100 * Mt_ML_load100  # t/GJ
LPC_ref = 3584.00  # €/t
annuity1 = (1-(1+discount_rate)**(-lifetime-1))/discount_rate
annuity2 = (1-(1+discount_rate)**(-lifetime/2-1))/discount_rate

# objective
model.ST_cost = (4.31) * 10 * (12293 + 365.02 * model.ST * 1) / 1e6  # M€
model.OPEX_part = (model.E + model.F + model.WT + model.H) / Mt_ML_load100 / 1e6 * \
            sum(model.production[t] * Mt_ML_load100 for t in model.N)  # M€
model.Revenue_part = LPC_ref * sum(model.production[t] * Mt_ML_load100 for t in model.N) / 1e6  # M€
model.CAPEX_part = (It_ML_ref[0]/annuity1 + model.ST_cost * (1 + 1 / 4.31 * 0.75)/annuity2) + OMt_ML_load100 + (1 + 0.2 + 0.6) * 0.06 * model.ST_cost  # M€
# maxmise annualised net profit
obj_expr = model.Revenue_part - (model.OPEX_part + model.CAPEX_part) * C6A_AF
model.objective = pe.Objective(sense=pe.maximize, expr=obj_expr)


# constraints
def con1(model, t):
    return model.M_li[t] * model.M_lo[t] == 0
model.con1 = pe.Constraint(model.N,rule=con1)

def con2(model,t):
    return 0 <= model.M_li[t] + model.production[t] * convert - model.M_lo[t]
model.con2 = pe.Constraint(model.N,rule=con2)

def con3(model,t):
    return model.M_li[t] + model.production[t] * convert - model.M_lo[t] <= load_100 * convert
model.con3 = pe.Constraint(model.N,rule=con3)

def con4(model,t):
    return model.M_li[t] + model.production[t] * convert - model.M_lo[t] <= model.power[t] * convert
model.con4 = pe.Constraint(model.N,rule=con4)

def con5(model,t):
    return 0 <= model.M_li[t]
model.con5 = pe.Constraint(model.N,rule=con5)

def con6(model,t):
    return model.M_li[t] <= model.ST
model.con6 = pe.Constraint(model.N,rule=con6)

def con7(model,t):
    return 0 <= model.M_lo[t]
model.con7 = pe.Constraint(model.N,rule=con7)

def con8(model,t):
    return model.M_lo[t] <= model.ST
model.con8 = pe.Constraint(model.N,rule=con8)

def Sum_rule2(model,t):
    return - model.ST <= sum((model.M_li[t2] - model.M_lo[t2]) for t2 in range(lower, t+1))
model.Sum_rule2 = pe.Constraint(model.N,rule=Sum_rule2)

def Sum_rule1(model,t):
    return sum((model.M_li[t2] - model.M_lo[t2]) for t2 in range(lower, t+1)) <= 0
model.Sum_rule1 = pe.Constraint(model.N,rule=Sum_rule1)

# one binary, one continuous
def con9(model, t):
    return model.production_select[t] * model.L <= model.production[t]
model.con9 = pe.Constraint(model.N,rule=con9)

def con10(model, t):
    return model.production[t] <= load_100 * model.production_select[t]
model.con10 = pe.Constraint(model.N,rule=con10)


# solve MIQCP
solver = po.SolverFactory('gurobi', solver_io='python')
solver.options['NonConvex'] = 2
solver.options['Presolve'] = 2
solver.options['NodefileStart'] = 1
solver.options["MIPGap"] = 0.001
solver.options["Cuts"] = 3
solver.options["MIPfocus"] = 3
solver.options["Method"] = 3
results = solver.solve(model, tee=True)

x_t = []
M_li_t = []
M_lo_t = []

for t in model.N:
    x_t.append(round(pe.value(model.production[t]), 2))
    M_li_t.append(round(pe.value(model.M_li[t]), 2))
    M_lo_t.append(round(pe.value(model.M_lo[t]), 2))
print(x_t) # production quantity profile, %
print(M_li_t) # intermediate chemicals storing profile, m3/h
print(M_lo_t) # intermediate chemicals usage profile, m3/h

print(pe.value(model.ST)) # optimal ST size
print(value(model.objective)) # highest ANP
